/****************************************************************************/
/* Ecmp 338, Assignment #6, Part III                                        */
/*      'bmsvr.c' - Contains all of the necessary buffer management         */
/*                   functions, for remote-calling purposes.                */
/* by Devon Smith (devon@ces.cwru.edu) and                                  */
/*    Andy Reitz (reitz@ces.cwru.edu)                                       */
/* Date: 4/26/98                                                            */
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <netdb.h>
#include <errno.h>
#include <signal.h>
#include "bm.h"

/*
 * The buffer is now done in FIFO order, using two global pointers. The
 * pointers will wrap around the array, and if at any given time the
 * empty index (eindex) points to an un-empty location, then the buffer is
 * full.
 */
int gbuffer[BUF_SIZE];	/* The Global Buffer-type repositiory. */
int dindex = 0;		/* Global Buffer index, pointing to next data entry. */
int eindex = 0;		/* Global Buffer index, pointing to next empty space. */
int semid = 0;		/* Global Semaphore Identifier. */

/* Global variables, definitions, etc. for Semaphores. */
union semun {
	int val;
	struct semid_ds *buf;
	ushort *array;
	};

/*
 * Global Variables.
 */
FILE *svr_log;		/* Server-wide log file, for keeping an eye on activity. */
void start_log_entry (void);

/*
 * This function should be executed when the server is killed!
 */
void sigterm_handler (int isignal)
{
	/* Close-up log file. */
	fclose (svr_log);

	/* Free up semaphores. */
	if (semctl (semid, 0, IPC_RMID, 0) == -1)
		{
		perror ("semctl -- BM_hndl -- semaphore free failed:");
		exit (2);
		}

	exit (0);
} /* End sigterm_handler(). */

/*
 * This function returns the next free element in the buffer. To do so,
 * it first gains sole access to the buffer (semaphore), then checks to
 * make sure that it's empty and within range. If so, it returns this
 * index. If not, it returns ERROR.
 */
int *elementindextoproduce_2_svc (void *argp, struct svc_req *rqstp)
{
	static int ret_index = 0;	/* The index to be returned. */

	/*
	 * These constructs make it easy to control who-is-in-command of
	 * the semaphore.
	 */
	struct sembuf acquire = {0, -1, SEM_UNDO},
	              release = {0,  1, SEM_UNDO};

	/*
	 * To start things off, acquire sole control of the buffer!
	 */
	acquire.sem_num = SEM_BUFFER;
	if (semop (semid, &acquire, 1) == ERROR)
		{
		start_log_entry();
		fprintf (svr_log, "semop -- EITP -- acquire buffer:");
		fprintf (svr_log, strerror(errno));
		fprintf (svr_log, "\n");
		kill (getpid(), SIGTERM);
		}

	/*
	 * Here is the core of this function. The following logic guides
	 * eindex selection. Basically, we start off by range-checking it.
	 * We assume that the index is always pointing to an _empty_
	 * position in the array, except for when the buffer is full.
	 */
	if ((eindex < BUF_SIZE) && (eindex >= 0))
		{
		if (gbuffer[eindex] != EMPTY)
			{
			/* Whelp, doesn't look very empty to me. Error out. */
			release.sem_num = SEM_BUFFER;
			if (semop (semid, &release, 1) == ERROR)
				{
				start_log_entry();
				fprintf (svr_log, "semop -- EITP -- release err1 buffer:");
				fprintf (svr_log, strerror(errno));
				fprintf (svr_log, "\n");
				kill (getpid(), SIGTERM);
				}

			start_log_entry();
			fprintf (svr_log, "EITP: No free space found in buffer.\n");

			ret_index = ERROR;
			return (&ret_index);
			} /* if gbuffer[index] != EMPTY */

		/*
		 * Things are looking good, increment the empty index for
		 * the next time.
		 */
		ret_index = eindex;
		eindex = (eindex + 1) % BUF_SIZE;

		start_log_entry();
		fprintf (svr_log, "EITP: free index %d found.\n", ret_index);

		/*
		 * Release the semaphore.
		 */
		release.sem_num = SEM_BUFFER;
		if (semop (semid, &release, 1) == ERROR)
			{
			start_log_entry();
			fprintf (svr_log, "semop -- EITP -- release buffer:");
			fprintf (svr_log, strerror(errno));
			fprintf (svr_log, "\n");
			kill (getpid(), SIGTERM);
			}

		/*
		 * Return the value to the empty index.
		 */
		return (&ret_index);
		}

	/*
	 * If we made it here, then it's an error. Release the semaphore,
	 * and say so.
	 */
	release.sem_num = SEM_BUFFER;
	if (semop (semid, &release, 1) == ERROR)
		{
		start_log_entry();
		fprintf (svr_log, "semop -- EITP -- release err2 buffer:");
		fprintf (svr_log, strerror(errno));
		fprintf (svr_log, "\n");
		kill (getpid(), SIGTERM);
		}

	ret_index = ERROR;
	return (&ret_index);
} /* End ElementIndexToProduce(). */

/*
 * This function takes an index and an input as parameters, and then based
 * on this, performs an insertion into the buffer. In order to do so,
 * however, it must first gain control of the buffer via semaphores. If
 * the insertion is successful, then this function returns TRUE, otherwise
 * it returns ERROR.
 */
int *producebufferelement_2_svc (int *argp, struct svc_req *rqstp)
{
	static int result;	/* Either TRUE or FALSE, eh? */

	/*
	 * These constructs make it easy to control who-is-in-command of
	 * the semaphore.
	 */
	struct sembuf acquire = {0, -1, SEM_UNDO},
	              release = {0,  1, SEM_UNDO};

	/*
	 * To start things off, acquire sole control of the buffer!
	 */
	acquire.sem_num = SEM_BUFFER;
	if (semop (semid, &acquire, 1) == ERROR)
		{
		start_log_entry();
		fprintf (svr_log, "semop -- PBE -- acquire buffer:");
		fprintf (svr_log, strerror(errno));
		fprintf (svr_log, "\n");
		kill (getpid(), SIGTERM);
		}

	/*
	 * This is the meat of this function. We check to see that we're
	 * given _both_ an index to an empty location in the buffer, and
	 * that our index doesn't leave the array (should never happen,
	 * so the check is just there for paranoia/security purposes).
	 */
	if ((gbuffer[argp[INDEX]] == EMPTY) && (argp[INDEX] < BUF_SIZE))
		{
		/* Looks good, insert the blasted thing. */
		gbuffer[argp[INDEX]] = argp[INPUT];

		start_log_entry();
		fprintf (svr_log, "PBE: %d inserted into location %d.\n", argp[INPUT], argp[INDEX]);

		/*
		 * Release the semaphore.
		 */
		release.sem_num = SEM_BUFFER;
		if (semop (semid, &release, 1) == ERROR)
			{
			start_log_entry();
			fprintf (svr_log, "semop -- PBE -- release buffer:");
			fprintf (svr_log, strerror(errno));
			fprintf (svr_log, "\n");
			kill (getpid(), SIGTERM);
			}

		result = TRUE;
		return (&result);
		}

	/*
	 * If we made it here, then it's an error. Release the semaphore,
	 * and say so.
	 */
	release.sem_num = SEM_BUFFER;
	if (semop (semid, &release, 1) == ERROR)
		{
		start_log_entry();
		fprintf (svr_log, "semop -- PBE -- release err buffer:");
		fprintf (svr_log, strerror(errno));
		fprintf (svr_log, "\n");
		kill (getpid(), SIGTERM);
		}

	fprintf (svr_log, "PBE: received invalid index.\n");

	result = ERROR;
	return (&result);
} /* End ProduceBufferElement(). */

/*
 * This function is supposed to return the next element from the managed
 * buffer, if such a beast exists. To do this, it must first gain sole
 * control of the buffer via semaphores, then it saves the data that
 * the index is pointing too, marks that space as being EMPTY, and then
 * increments the data index pointer. If for any reason the index is
 * out-of-range, the function returns ERROR.
 */
int *consumerbufferelement_2_svc (void *argp, struct svc_req *rqstp)
{
	/* The data to be eaten.... */
	static int consumed = 0;

	/*
	 * These constructs make it easy to control who-is-in-command of
	 * the semaphore.
	 */
	struct sembuf acquire = {0, -1, SEM_UNDO},
	              release = {0,  1, SEM_UNDO};

	/*
	 * To start things off, acquire sole control of the buffer!
	 */
	acquire.sem_num = SEM_BUFFER;
	if (semop (semid, &acquire, 1) == ERROR)
		{
		start_log_entry();
		fprintf (svr_log, "semop -- CBE -- acquire buffer:");
		fprintf (svr_log, strerror(errno));
		fprintf (svr_log, "\n");
		kill (getpid(), SIGTERM);
		}

	/* So long as the current data index isn't EMPTY, we're ok. */
	if (gbuffer[dindex] != EMPTY)
		{
		consumed = gbuffer[dindex];

		start_log_entry();
		fprintf (svr_log, "CBE: %d consumed from index %d.\n", consumed, dindex);

		gbuffer[dindex] = EMPTY;
		dindex = (dindex + 1) % BUF_SIZE;

		/*
		 * Release the semaphore, return the consumed.
		 */
		release.sem_num = SEM_BUFFER;
		if (semop (semid, &release, 1) == ERROR)
			{
			start_log_entry();
			fprintf (svr_log, "semop -- CBE -- release buffer:");
			fprintf (svr_log, strerror(errno));
			fprintf (svr_log, "\n");
			kill (getpid(), SIGTERM);
			}

		return (&consumed);
		}

	/*
	 * If we made it here, then it's an error. Release the semaphore,
	 * and say so.
	 */
	release.sem_num = SEM_BUFFER;
	if (semop (semid, &release, 1) == ERROR)
		{
		start_log_entry();
		fprintf (svr_log, "semop -- CBE -- release err buffer:");
		fprintf (svr_log, strerror(errno));
		fprintf (svr_log, "\n");
		kill (getpid(), SIGTERM);
		}

	start_log_entry();
	fprintf (svr_log, "CBE: Consumer tried to take from empty buffer location %d.\n", dindex);

	consumed = ERROR;
	return (&consumed);
} /* End ConsumerBufferElement(). */

/*
 * This routine initializes all of the necessary structures needed for
 * the buffer management process to run. Normally, this sort of thing
 * would be done in main(), but I really don't feel like hacking-up
 * rpcgen-generated code. So, this routine will be called both by the
 * consumer *and* producer functions, but it has a check so that it will
 * only execute the initialization code once.
 */
int *initializebuffer_2_svc (void *argp, struct svc_req *rqstp)
{
	static int counter = 0, result;
	char fname[30];

	ushort start_val[NUM_SEM];	/* Initial values for Semaphore. */
	union semun ctl_arg;		/* Argument union for semctl(). */
	key_t sem_key;

	/* Just so we can print pretty in the logs. */
	time_t cur_time;
	char *conv_time;
	char hostname[MAXHOSTNAMELEN];

	/* Check to see if we've been run before! */
	if (counter > 0)
		{
		result = ERROR;
		return (&result);
		}

	/* Open the log file! */
	sprintf (fname, "/tmp/ajr_server_%d.out", getpid());
	if ((svr_log = fopen (fname, "a+")) == NULL)
		{
		/* Can't open the durn'd file. */
		fprintf (stderr, "Server: Cannot open log file '%s'.\n", fname);
		kill (getpid(), SIGTERM);
		}

	cur_time = time (NULL);
	conv_time = ctime (&cur_time);
	conv_time[strlen(conv_time) - 6] = '\0';
	gethostname (hostname, MAXHOSTNAMELEN);
	fprintf (svr_log, "Log file started on server %s, pid %d, at %s.\n", hostname, getpid(), conv_time);
	fprintf (svr_log, "   -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
	free (conv_time);

	/* Grab the program-quitting signals. */
	signal (SIGTERM, sigterm_handler);
	signal (SIGQUIT, sigterm_handler);
	signal (SIGINT, sigterm_handler);

	/*
	 * Allocate our *Semaphore*. This particular semaphore will exist of
	 * NUM_SEM locations. In this case, there will only be one:
	 *      1. The Buffer Semaphore (Controls access to the buffer,
	 *         so that the BM functions don't all try and modify it
	 *         concurrently.
	 */
	sem_key = ftok (".", 'S');
	if ((semid = semget (sem_key, NUM_SEM, IPC_CREAT | IPC_EXCL | 0600)) == -1)
		{
		start_log_entry();
		fprintf (svr_log, "semget -- BM -- create semaphore:");
		fprintf (svr_log, strerror(errno));
		fprintf (svr_log, "\n");
		kill (getpid(), SIGTERM);
		}

	/*
	 * Now, initialize this here semaphore, like so:
	 *      semaphore[SEM_BUFFER] = 1
	 */
	start_val[SEM_BUFFER] = 1;
	ctl_arg.array = start_val;

	if (semctl (semid, 0, SETALL, ctl_arg) == -1)
		{
		start_log_entry();
		fprintf (svr_log, "semctl -- BM -- SETALL failed:");
		fprintf (svr_log, strerror(errno));
		fprintf (svr_log, "\n");
		kill (getpid(), SIGTERM);
		}

	/* Initialize the buffer! */
	for (counter = 0; counter < BUF_SIZE; counter++)
		gbuffer[counter] = EMPTY;

	result = TRUE;
	return(&result);
} /* End InitializeBuffer(). */

/*
 * The code necessary to print out a log entry is pretty repetitive (all
 * of this time and hostname stuff), so this function just encapsulates it
 * all, and when called writes out a partial entry to 'svr_log'.
 */
void start_log_entry (void)
{
	/* Just so we can print pretty in the logs. */
	time_t cur_time;
	char *conv_time;
	char hostname[MAXHOSTNAMELEN];

	cur_time = time (NULL);
	conv_time = ctime (&cur_time);
	conv_time[strlen(conv_time) - 6] = '\0';	/* not pretty, but gets */
	gethostname (hostname, MAXHOSTNAMELEN);		/* rid of trailing 'year\n'. */

	fprintf (svr_log, "%s %s ", conv_time, hostname);

	free (conv_time);
} /* End start_log_entry(). */

